import { Config, loadConfig } from "./config";
import SMB from "@marsaud/smb2";
import mysql from "promise-mysql";
import moment from "moment";
import { Message, readChatBuffer } from "./parse-buffer";
import bunyan from "bunyan";

export class SMBReader {
	logPrefix: string;
	logLevel: bunyan.LogLevel;
	constructor(logPrefix: string, logLevel?: bunyan.LogLevel) {
		this.logPrefix = logPrefix;
		this.logLevel = logLevel;
	}
	config: Config;
	smb: SMB;
	db: mysql.Pool;
	log: bunyan;
	async init() {
		this.log = bunyan.createLogger({ name: this.logPrefix, level: this.logLevel || "info" });
		this.log.info(`Start initializing.`);
		this.log.debug(`Reading config...`);
		this.config = await loadConfig();
		this.smb = new SMB(this.config.smbConfig);
		this.log.debug(`Connecting to database...`);
		this.db = await mysql.createPool(this.config.MySQLConfig);
		this.log.debug(`Creating table...`);
		await this.db.query("CREATE TABLE IF NOT EXISTS `filesRead` (\n" +
			"  `fileid` bigint(20) UNSIGNED NOT NULL AUTO_INCREMENT,\n" +
			"  `date` datetime UNIQUE NOT NULL,\n" +
			"  `filename` varchar(40) COLLATE utf8_unicode_ci NOT NULL,\n" +
			"  PRIMARY KEY (fileid),\n" +
			"  UNIQUE KEY (date),\n" +
			"  UNIQUE KEY (filename)\n" +
			") ENGINE=InnoDB AUTO_INCREMENT=2 DEFAULT CHARSET=utf8 COLLATE=utf8_unicode_ci");
		await this.db.query("CREATE TABLE IF NOT EXISTS `messages` (\n" +
			"  `id` bigint(20) UNSIGNED NOT NULL AUTO_INCREMENT,\n" +
			"  `fileid` bigint(20) UNSIGNED NOT NULL,\n" +
			"  `channel` varchar(4) COLLATE utf8_unicode_ci NOT NULL,\n" +
			"  `speaker` varchar(16) COLLATE utf8_unicode_ci NOT NULL,\n" +
			"  `content` text COLLATE utf8_unicode_ci NOT NULL,\n" +
			"  PRIMARY KEY (id),\n" +
			"  FOREIGN KEY (fileid) REFERENCES filesRead(fileid),\n" +
			"  KEY (channel(4)),\n" +
			"  INDEX (speaker(16))\n" +
			") ENGINE=InnoDB AUTO_INCREMENT=2 DEFAULT CHARSET=utf8 COLLATE=utf8_unicode_ci");
		this.log.info(`Initialization finished.`);
	}
	private getFilePath(filename: string) {
		return this.config.pathPrefix ? `${this.config.pathPrefix}\\${filename}` : filename;
	}
	private flatternMessages(messages: Message[]): string {
		return messages.map(m => JSON.stringify(m)).join("|");
	}
	async run() {
		this.log.info(`Started.`);
		const db = await this.db.getConnection();
		try {
			this.log.info(`Reading file list...`);
			let filelist = (await this.smb.readdir(this.config.pathPrefix)).filter(m => m.match(/^chat_.*\.html$/));
			const files = filelist.map(filename => {
				const dateString = filename.match(/^chat_(.*)\.html$/)[1].replace(/-/g, ":");
				const date = moment(dateString).utcOffset("+08:00", true)
				return {
					name: filename,
					date
				}
			});
			files.sort((f1, f2) => {
				return f1.date.unix() - f2.date.unix();
			});
			for (let file of files) {
				const filePath = this.getFilePath(file.name);
				try {
					const testIndex: any[] = await db.query("select * from `filesRead` where filename = ?", [file.name]);
					if (testIndex.length) {
						this.log.debug(`File ${filePath} has already been read. Skipping.`);
						continue;
					}
					await db.beginTransaction();
					this.log.info(`Reading file ${filePath}...`);
					const fileMetadata = await db.query("insert into `filesRead` set ?", {
						filename: file.name,
						date: file.date.format("YYYY-MM-DD HH:mm:ss")
					});
					const insertedFileID: number = fileMetadata.insertId;
					const previousFilenameArray: any[] = await db.query("select filename from `filesRead` where fileid = ?", [insertedFileID - 1]);
					let currentMessages = readChatBuffer((await this.smb.readFile(filePath)) as Buffer);
					if (previousFilenameArray.length && files.find(f => f.name === previousFilenameArray[0].filename)) { // compare with old file
						const previousFilePath = this.getFilePath(previousFilenameArray[0].filename);
						this.log.debug(`Comparing file ${filePath} with old file ${previousFilePath}.`)
						try {
							const previousMessages = readChatBuffer((await this.smb.readFile(previousFilePath)) as Buffer);
							const flattenedPreviousMessages = this.flatternMessages(previousMessages);
							const flattenedCurrentMessages = this.flatternMessages(currentMessages);
							if (flattenedCurrentMessages.startsWith(flattenedPreviousMessages)) {
								const sliceCount = previousMessages.length;
								this.log.debug(`Slicing the first ${sliceCount} messages of file ${filePath}.`);
								currentMessages = currentMessages.slice(sliceCount);
							}
						} catch (e) {
							this.log.warn(`Failed comparing file ${filePath} with old file ${previousFilePath}. Skipping:`, e.toString());
						}
					}
					for (let message of currentMessages) {
						this.log.debug(`Recording message from file ${insertedFileID} ${filePath}:`, JSON.stringify(message));
						await db.query("insert into `messages` set ?", {
							fileid: insertedFileID,
							...message
						});
					}
					const deleteFilenameArray: any[] = await db.query("select filename from `filesRead` where fileid = ?", [insertedFileID - 2]);
					if (deleteFilenameArray.length && files.find(f => f.name === deleteFilenameArray[0].filename)) { // delete file with id - 2 because useless
						const deleteFilePath = this.getFilePath(deleteFilenameArray[0].filename);
						try {
							this.log.debug(`Deleting file ${deleteFilePath}.`);
							await this.smb.unlink(deleteFilePath);
						} catch (e) {
							this.log.debug(`Failed deleting file ${deleteFilePath}:`, e.toString());
						}
					}
					await db.commit();
				} catch (e) {
					await db.rollback();
					this.log.warn(`Errored reading file ${filePath}:`, e.toString());
				}
			}
		} catch (e) {
			this.log.warn(`Read error:`, e.toString());
		}
		db.release();
		this.log.info(`Finished.`);
	}
}
